use std::ops;
use anyhow::*;
use percent_encoding::percent_decode_str;
use uriparse::URIReference;
use rustls::Certificate;
#[cfg(feature="user_management")]
use serde::{Serialize, de::DeserializeOwned};

#[cfg(feature="user_management")]
use crate::user_management::{UserManager, User};

pub struct Request {
    uri: URIReference<'static>,
    input: Option<String>,
    certificate: Option<Certificate>,
    #[cfg(feature="user_management")]
    manager: UserManager,
}

impl Request {
    pub fn from_uri(
        uri: URIReference<'static>,
        #[cfg(feature="user_management")]
        manager: UserManager,
    ) -> Result<Self> {
        Self::with_certificate(
            uri,
            None,
            #[cfg(feature="user_management")]
            manager
        )
    }

    pub fn with_certificate(
        mut uri: URIReference<'static>,
        certificate: Option<Certificate>,
        #[cfg(feature="user_management")]
        manager: UserManager,
    ) -> Result<Self> {
        uri.normalize();

        let input = match uri.query() {
            None => None,
            Some(query) => {
                let input = percent_decode_str(query.as_str())
                    .decode_utf8()
                    .context("Request URI query contains invalid UTF-8")?
                    .into_owned();
                Some(input)
            }
        };

        Ok(Self {
            uri,
            input,
            certificate,
            #[cfg(feature="user_management")]
            manager,
        })
    }

    pub fn uri(&self) -> &URIReference {
        &self.uri
    }

    pub fn path_segments(&self) -> Vec<String> {
        self.uri()
            .path()
            .segments()
            .iter()
            .map(|segment| percent_decode_str(segment.as_str()).decode_utf8_lossy().into_owned())
            .collect::<Vec<String>>()
    }

    pub fn input(&self) -> Option<&str> {
        self.input.as_deref()
    }

    pub fn set_cert(&mut self, cert: Option<Certificate>) {
        self.certificate = cert;
    }

    pub fn certificate(&self) -> Option<&Certificate> {
        self.certificate.as_ref()
    }

    #[cfg(feature="user_management")]
    /// Attempt to determine the user who sent this request
    ///
    /// May return a variant depending on if the client used a client certificate, and if
    /// they've registered as a user yet.
    pub fn user<UserData>(&self) -> Result<User<UserData>>
    where
        UserData: Serialize + DeserializeOwned
    {
        Ok(self.manager.get_user(self.certificate())?)
    }
}

impl ops::Deref for Request {
    type Target = URIReference<'static>;

    fn deref(&self) -> &Self::Target {
        &self.uri
    }
}
